// vector3 class
// -Kyoung Shin Park (2011 Spring)
//////////////////////////////////////////////////////////////////////
#ifndef __VECTOR3_H__
#define __VECTOR3_H__

#include <iostream>
using namespace std;

namespace oglclass
{

class vector3
{
public:
    static vector3 Zero;

	// constructor
	vector3(void) { v_[0] = v_[1] = v_[2] = 0.0f; }
    vector3(const vector3 &other);
	vector3(const float *other);
    vector3(float x, float y, float z);
    void set(float x, float y, float z);

	// length & normalize
    float length(void);
    void normalize(void);

	// static utility methods
    static float distance(const vector3 &v1, const vector3 &v2);
    static float dotProduct(const vector3 &v1,  const vector3 &v2 );
    static vector3 crossProduct(const vector3 &v1, const vector3 &v2);
    static float theta(const vector3 &v1, const vector3 &v2);

    // assignment operators...
    vector3& operator = (const vector3 &other);
    vector3& operator += (const vector3 &other);
    vector3& operator -= (const vector3 &other);
    vector3& operator *= (const float scalar);
    vector3& operator /= (const float scalar);

	// equality operators
	bool operator == (const vector3 &other) const;
	bool operator != (const vector3 &other) const;

	// vector operators
    vector3 operator + (const vector3 &other);
    vector3 operator - (const vector3 &other);
    vector3 operator * (const vector3 &other);
    vector3 operator / (const vector3 &other);

    vector3 operator * (const float scalar);
    friend vector3 operator * (const float scalar, const vector3 &other);
    vector3 operator / (const float scalar);
    friend vector3 operator / (const float scalar, const vector3 &other);

    vector3 operator + (void) const;
    vector3 operator - (void) const;

	float operator[] (const int i) const;
	float& operator[] (const int i);
	operator float* () { return v_; } // type conversion operator overloading

	friend ostream& operator<<(ostream& s, const vector3& v);

protected:
    // member variable
    float v_[3];
};

inline ostream& operator<< (ostream& s, const vector3& v)
{ 
	return s << "(" << v[0] << ", " << v[1] << ", " << v[2] << ")";
}

} // namespace

#endif
